@@ -6,5 +6,5 @@
 /* in the file PATENTS.  All contributing project authors may */
 /* be found in the AUTHORS file in the root of the source tree. */
 #include "vpx/vpx_codec.h"
-static const char* const cfg = "--prefix=/usr --enable-vp8 --enable-vp9 --enable-vp9-highbitdepth --enable-vp9-postproc --enable-vp9-temporal-denoising --enable-webm-io --size-limit=16384x16384";
+static const char* const cfg = "--prefix=/usr --enable-pic --enable-vp8 --enable-vp9 --enable-vp9-highbitdepth --enable-vp9-postproc --enable-vp9-temporal-denoising --enable-webm-io --size-limit=16384x16384";
 const char *vpx_codec_build_config(void) {return cfg;}
